from flask_authx.entities import Session, User
from flask_authx.forms import SessionForm, UserForm
from flask_authx.errors import (
    ConflictError,
    DatabaseError,
    InvalidCredentialsError,
    NotAuthenticatedError,
    NotFoundError,
)
from flask_authx.interfaces.repository import ISessionsRepository
from flask_authx.interfaces.service import IAuthService, IUsersService
from flask_authx.interfaces.security import IPasswordHashing
from flask_authx.security.tokens import generate_access_token
from flask_authx.utils.result import Result


class AuthService(IAuthService):
    def __init__(
        self,
        sessions_repository: ISessionsRepository,
        users_service: IUsersService,
        password_hashing: IPasswordHashing,
    ) -> None:
        self.sessions_db = sessions_repository
        self.users_service = users_service
        self.password_hashing = password_hashing

    def login(
        self, u: UserForm
    ) -> Result[
        Session,
        InvalidCredentialsError | ConflictError | DatabaseError,
    ]:
        user_res = self.users_service.get_by_name(u.username)

        if not user_res.success:
            return Result.fail(InvalidCredentialsError("Invalid credentials"))

        user = user_res.value  # Contains hashed password

        if not self.password_hashing.verify(u.password, user.password):
            return Result.fail(InvalidCredentialsError("Invalid credentials"))

        upd_auth_res = self.users_service.mark_authenticated(user.id)

        if not upd_auth_res.success:
            if not upd_auth_res.error_is(NotFoundError):
                return Result.fail(NotFoundError(f"User with id {id} not found"))
            return Result.fail(DatabaseError(f"Cannot update the state of the user: INFO {upd_auth_res.error.message}"))

        new_session = SessionForm(token=generate_access_token(), user=user)
        new_sess_res = self.sessions_db.add(new_session)

        if not new_sess_res.success:
            if new_sess_res.error_is(ConflictError):
                return Result.fail(ConflictError("User is already logged in"))

            else:
                return Result.fail(DatabaseError("Cannot create the new session"))

        return Result.ok(new_sess_res.value)

    def logout(self, user: User) -> Result[None, NotAuthenticatedError | DatabaseError]:
        sess_res = self.sessions_db.get_by_user(user.id)

        if not sess_res.success:
            if sess_res.error_is(NotFoundError):
                return Result.fail(
                    NotAuthenticatedError(f"No session linked to user {user.id}")
                )

            else:
                return Result.fail(DatabaseError("Cannot find the session"))

        sess = sess_res.value

        rem_sess_res = self.sessions_db.remove(sess.token)

        if not rem_sess_res.success:
            return Result.fail(DatabaseError(rem_sess_res.error.message))

        upd_usr_res = self.users_service.mark_unauthenticated(sess.user.id)

        if not upd_usr_res.success:
            return Result.fail(DatabaseError(upd_usr_res.error.message))

        return Result.ok(None)



from abc import ABC, abstractmethod

from flask_authx.entities import Session, User
from flask_authx.role import Role
from flask_authx.forms import UserForm
from flask_authx.errors import (
    ConflictError,
    InvalidCredentialsError,
    NotAuthenticatedError,
    DatabaseError,
    NotFoundError,
    ValidationError,
)
from flask_authx.utils.result import Result


class IAuthService(ABC):
    @abstractmethod
    def login(
        self, u: UserForm
    ) -> Result[
        Session,
        InvalidCredentialsError | ConflictError | DatabaseError,
    ]:
        pass

    @abstractmethod
    def logout(self, user: User) -> Result[None, NotAuthenticatedError | DatabaseError]:
        pass


class IUsersService(ABC):
    @abstractmethod
    def all(self) -> Result[list[User], DatabaseError]:
        pass

    @abstractmethod
    def get_by_id(self, id: int) -> Result[User, NotFoundError | DatabaseError]:
        pass

    @abstractmethod
    def get_by_name(self, username: str) -> Result[User, NotFoundError | DatabaseError]:
        pass

    @abstractmethod
    def add(
        self, form: UserForm
    ) -> Result[User, ConflictError | ValidationError | DatabaseError]:
        pass

    @abstractmethod
    def remove(self, id: int) -> Result[None, NotFoundError | DatabaseError]:
        pass

    @abstractmethod
    def update_name(
        self, id: int, new_username: str
    ) -> Result[User, ValidationError | NotFoundError | ConflictError | DatabaseError]:
        pass

    @abstractmethod
    def update_role(
        self, id: int, role: Role
    ) -> Result[User, NotFoundError | DatabaseError]:
        pass

    @abstractmethod
    def mark_authenticated(
        self, id: int
    ) -> Result[User, NotFoundError | DatabaseError]:
        pass

    @abstractmethod
    def mark_unauthenticated(
        self, id: int
    ) -> Result[User, NotFoundError | DatabaseError]:
        pass

    @abstractmethod
    def is_authenticated(self, id: int) -> Result[bool, NotFoundError | DatabaseError]:
        pass



import bcrypt

from flask_authx.interfaces.security import IPasswordHashing


class BcryptPasswordHashing(IPasswordHashing):
    def hash(self, password: str) -> str:
        salt = bcrypt.gensalt(12)
        return bcrypt.hashpw(password.encode("utf-8"), salt).decode("utf-8")

    def verify(self, password: str, hashed: str) -> bool:
        return bcrypt.checkpw(password.encode("utf-8"), hashed.encode("utf-8"))



from http import HTTPStatus
from typing import Optional

from flask import make_response

from flask_authx.entities import Session, User


class UserNotFoundResponse:
    @staticmethod
    def create(*, id: Optional[int] = None, username: Optional[str] = None):
        if id is not None:
            message = f"User with id {id} not found."
        elif username is not None:
            message = f"User with username: {username} not found."
        else:
            message = "User not found."

        res = make_response({"success": False, "error": message})
        res.status_code = HTTPStatus.NOT_FOUND
        return res


class InternalErrorResponse:
    @staticmethod
    def create(message: str):
        res = make_response({"success": False, "error": message})
        res.status_code = HTTPStatus.INTERNAL_SERVER_ERROR
        return res


class UserInfoResponse:
    @staticmethod
    def create(user: User):
        res = make_response(
            {
                "success": True,
                "user": {
                    "id": user.id,
                    "username": user.username,
                    "role": user.role.dump(),
                    "is_authenticated": user.is_authenticated,
                    "created_at": user.created_at.isoformat(),
                },
            }
        )
        res.status_code = HTTPStatus.OK
        return res


class UsersListResponse:
    @staticmethod
    def create(users: list[User]):
        res = make_response(
            {
                "success": True,
                "users": [
                    {"id": u.id, "username": u.username, "role": u.role.dump()}
                    for u in users
                ],
            }
        )
        res.status_code = HTTPStatus.OK
        return res


class SuccessfulEmptyResponse:
    @staticmethod
    def create():
        res = make_response({"success": True})
        res.status_code = HTTPStatus.OK
        return res


class JSONRequiredResponse:
    @staticmethod
    def create():
        res = make_response(
            {
                "success": False,
                "error": "MimeType must be 'application/json'",
            }
        )
        res.status_code = HTTPStatus.UNSUPPORTED_MEDIA_TYPE
        return res


class MissingJsonFieldsResponse:
    @staticmethod
    def create(*fields: str):
        res = make_response(
            {
                "success": False,
                "error": f"Missing fields in JSON body: {', '.join([f"'{f}'" for f in fields])}",
            }
        )
        res.status_code = HTTPStatus.BAD_REQUEST
        return res


class CreatedUserResponse:
    @staticmethod
    def create(user: User, location: str):
        res = make_response(UserInfoResponse.create(user))
        res.status_code = HTTPStatus.CREATED
        res.headers.set("Location", location)
        return res


class BadRequestFormatResponse:
    @staticmethod
    def create(message: str):
        res = make_response({"success": False, "error": message})
        res.status_code = HTTPStatus.BAD_REQUEST
        return res


class UsernameConflictResponse:
    @staticmethod
    def create():
        res = make_response(
            {
                "success": False,
                "error": "That username is already taken.",
            }
        )
        res.status_code = HTTPStatus.CONFLICT
        return res


class UserFormValidationErrorResponse:
    @staticmethod
    def create(message: Optional[str] = None):
        res = make_response(
            {
                "success": False,
                "error": message or "Invalid credentials.",
            }
        )
        res.status_code = HTTPStatus.BAD_REQUEST
        return res


class RequireAuthenticationResponse:
    @staticmethod
    def create():
        res = make_response(
            {
                "success": False,
                "error": "The requested action requires authentication.",
            }
        )
        res.status_code = HTTPStatus.UNAUTHORIZED
        return res


class MissingQueryParamsResponse:
    @staticmethod
    def create(*fields: str):
        res = make_response(
            {
                "success": False,
                "error": f"Missing params in query string: {', '.join([f"'{f}'" for f in fields])}",
            }
        )
        res.status_code = HTTPStatus.BAD_REQUEST
        return res


class PermissionErrorResponse:
    @staticmethod
    def create():
        res = make_response(
            {
                "success": False,
                "error": "Your account doesn't have the required permission to proceed with the requested action.",
            }
        )
        res.status_code = HTTPStatus.FORBIDDEN
        return res


class InvalidCredentialsResponse:
    @staticmethod
    def create():
        res = make_response(
            {
                "success": False,
                "error": "Invalid username or password.",
            }
        )
        res.status_code = HTTPStatus.UNAUTHORIZED
        return res


class SessionAlreadyOpenResponse:
    @staticmethod
    def create():
        res = make_response(
            {
                "success": False,
                "error": "Session already open on another device. Please log out from there first.",
            }
        )
        res.status_code = HTTPStatus.UNAUTHORIZED
        return res


class SuccessfulLoginResponse:
    @staticmethod
    def create(session: Session):
        res = make_response({"success": True, "access_token": session.token})
        res.status_code = HTTPStatus.OK
        return res



from flask_authx.forms import SessionForm
from flask_authx.entities import Session
from flask_authx.errors import (
    ConflictError,
    DatabaseError,
    NotFoundError,
    ValidationError,
)
from flask_authx.database.sqlalchemy.shared.instance import get_instance
from flask_authx.database.sqlalchemy.models import get_session_model
from flask_authx.database.sqlalchemy.shared.errors import (
    handle_database_error,
    handle_integrity_error,
)
from flask_authx.interfaces.repository import ISessionsRepository
from flask_authx.utils.result import Result


class SQLAlchemySessionsRepository(ISessionsRepository):
    def __init__(self) -> None:
        super().__init__()
        self.SessionModel = get_session_model()

    @handle_database_error
    def all(self) -> Result[list[Session], DatabaseError]:
        sessions = get_instance().session.query(self.SessionModel).all()
        return Result.ok([s.to_entity() for s in sessions])

    @handle_database_error
    def get_by_token(
        self, token: str
    ) -> Result[Session, NotFoundError | DatabaseError]:
        sess = get_instance().session.get(self.SessionModel, token)
        if not sess:
            return Result.fail(
                NotFoundError(f"Session with token: {token} doesn't exist.")
            )
        return Result.ok(sess.to_entity())

    @handle_database_error
    def get_by_user(
        self, user_id: int
    ) -> Result[Session, NotFoundError | DatabaseError]:
        sess = (
            get_instance()
            .session.query(self.SessionModel)
            .filter(self.SessionModel.user_id == user_id)
            .first()
        )

        if not sess:
            return Result.fail(
                NotFoundError(f"Session for user with id: {user_id} doesn't exist.")
            )

        return Result.ok(sess.to_entity())

    @handle_database_error
    @handle_integrity_error
    def add(
        self, new: SessionForm
    ) -> Result[Session, ConflictError | ValidationError | DatabaseError]:
        model = self.SessionModel.create(new)
        get_instance().session.add(model)
        get_instance().session.commit()
        return Result.ok(model.to_entity())

    @handle_database_error
    def remove(self, token: str) -> Result[None, NotFoundError | DatabaseError]:
        model = get_instance().session.get(self.SessionModel, token)
        if not model:
            return Result.fail(
                NotFoundError(f"Session with token: {token} doesn't exist.")
            )
        get_instance().session.delete(model)
        get_instance().session.commit()
        return Result.ok(None)

    @handle_database_error
    def remove_by_user(
        self, user_id: int
    ) -> Result[None, NotFoundError | DatabaseError]:
        model = (
            get_instance()
            .session.query(self.SessionModel)
            .filter(self.SessionModel.user_id == user_id)
            .first()
        )

        if not model:
            return Result.fail(
                NotFoundError(f"Session for user with id: {user_id} doesn't exist.")
            )

        get_instance().session.delete(model)
        get_instance().session.commit()
        return Result.ok(None)



from abc import ABC, abstractmethod


class IPasswordHashing(ABC):
    @abstractmethod
    def hash(self, password: str) -> str:
        pass

    @abstractmethod
    def verify(self, password: str, hashed: str) -> bool:
        pass



from abc import ABC, abstractmethod
from typing import Any

from flask_authx.entities import Session, User
from flask_authx.forms import SessionForm, UserForm, UserUpdateData
from flask_authx.errors import (
    ConflictError,
    DatabaseError,
    NotFoundError,
    ValidationError,
)
from flask_authx.utils.result import Result


class IUsersRepository(ABC):
    @abstractmethod
    def all(self) -> Result[list[User], DatabaseError]:
        pass

    @abstractmethod
    def get_by_id(self, id: int) -> Result[User, NotFoundError | DatabaseError]:
        pass

    @abstractmethod
    def get_by_name(self, username: str) -> Result[User, NotFoundError | DatabaseError]:
        pass

    @abstractmethod
    def add(
        self, new: UserForm
    ) -> Result[User, ConflictError | ValidationError | DatabaseError]:
        pass

    @abstractmethod
    def remove(self, id: int) -> Result[None, NotFoundError | DatabaseError]:
        pass

    @abstractmethod
    def update(
        self, id: int, values: UserUpdateData
    ) -> Result[User, ConflictError | ValidationError | NotFoundError | DatabaseError]:
        pass


class ISessionsRepository(ABC):
    @abstractmethod
    def all(self) -> Result[list[Session], DatabaseError]:
        pass

    @abstractmethod
    def get_by_token(
        self, token: str
    ) -> Result[Session, NotFoundError | DatabaseError]:
        pass

    @abstractmethod
    def get_by_user(
        self, user_id: int
    ) -> Result[Session, NotFoundError | DatabaseError]:
        pass

    @abstractmethod
    def add(
        self, new: SessionForm
    ) -> Result[Session, ConflictError | ValidationError | DatabaseError]:
        pass

    @abstractmethod
    def remove(self, token: str) -> Result[None, NotFoundError | DatabaseError]:
        pass

    @abstractmethod
    def remove_by_user(
        self, user_id: int
    ) -> Result[None, NotFoundError | DatabaseError]:
        pass



from flask import Flask
from flask_sqlalchemy import SQLAlchemy

from flask_authx.errors import (
    AppError,
    DatabaseError,
    NotFoundError,
    ConfigurationError,
)
from flask_authx.forms import UserForm
from flask_authx.role import Role
from flask_authx.interfaces.database import IDatabaseSetup
from flask_authx.interfaces.service import IUsersService
from flask_authx.config import FIRST_USER_NAME, FIRST_USER_PASSWORD


class SQLAlchemyDatabaseSetup(IDatabaseSetup):
    def __init__(self, app: Flask, users_service: IUsersService) -> None:
        self.app = app
        self.sqlalchemy: SQLAlchemy = app.extensions["sqlalchemy"]
        self.users_service = users_service

    def init(self) -> None:
        with self.app.app_context():
            # Import and create models dynamically for this SQLAlchemy instance
            from flask_authx.database.sqlalchemy.models import (
                get_user_model,
                get_session_model,
            )

            get_user_model()
            get_session_model()

            if not self.app.extensions.get("migrate"):
                self.sqlalchemy.create_all()

            res = self.users_service.get_by_name(FIRST_USER_NAME)

            if not res.success:
                if res.error_is(NotFoundError):
                    first_user = UserForm(
                        username=FIRST_USER_NAME,
                        password=FIRST_USER_PASSWORD,
                        role=Role.admin,
                    )

                    add_res = self.users_service.add(first_user)

                    if not add_res.success:
                        raise DatabaseError(
                            (
                                "It was not possible to create the first "
                                f"administrator user. Error: {add_res.error.message}"
                            )
                        )

                elif res.error_is(DatabaseError):
                    raise DatabaseError(res.error.message)

                else:
                    raise AppError(f"Unexpected error: {res.error}")



class ProgrammingError(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class AppError(Exception):
    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(message)


class ConfigurationError(AppError):
    pass


class NotFoundError(AppError):
    pass


class ValidationError(AppError):
    pass


class DatabaseError(AppError):
    pass


class ConflictError(AppError):
    pass


class DependentResourceError(AppError):
    pass


class NotAuthenticatedError(AppError):
    pass


class InvalidCredentialsError(AppError):
    pass



from dataclasses import dataclass
from typing import TypedDict

from flask_authx.entities import User
from flask_authx.errors import ValidationError
from flask_authx.role import Role
from flask_authx.utils.result import Result


@dataclass(frozen=True)
class UserForm:
    username: str
    password: str
    role: Role

    def is_valid(self) -> Result[None, ValidationError]:
        rules = [
            User.validate_username(self.username),
            User.validate_password(self.password),
        ]

        for r in rules:
            if not r.success:
                return r

        return Result.ok(None)


@dataclass
class SessionForm:
    token: str
    user: User


class UserUpdateData(TypedDict, total=False):
    username: str
    role: Role
    is_authenticated: bool



import sqlalchemy.exc as sae

from flask_authx.errors import (
    ProgrammingError,
    DatabaseError,
    ConflictError,
    ValidationError,
)
from flask_authx.utils.result import Result
from .instance import instance


def handle_database_error(f):
    def wrapper(*args, **kwargs):
        db_session = instance.session
        try:
            return f(*args, **kwargs)

        except sae.ProgrammingError as ex:
            db_session.rollback()
            raise ProgrammingError(str(ex))

        except sae.SQLAlchemyError as ex:
            db_session.rollback()
            return Result.fail(DatabaseError(str(ex)))

    return wrapper


def handle_integrity_error(f):
    def wrapper(*args, **kwargs):
        db_session = instance.session
        try:
            return f(*args, **kwargs)

        except sae.IntegrityError as ex:
            if "unique" in str(ex).lower():
                return Result.fail(
                    ConflictError("Two rows cannot have the same identificator.")
                )
            db_session.rollback()
            return Result.fail(ValidationError(str(ex)))

    return wrapper



from flask_authx.entities import User
from flask_authx.forms import UserForm, UserUpdateData
from flask_authx.errors import (
    DatabaseError,
    NotFoundError,
    ConflictError,
    ValidationError,
)
from flask_authx.database.sqlalchemy.shared.instance import get_instance
from flask_authx.database.sqlalchemy.models import get_user_model
from flask_authx.database.sqlalchemy.shared.errors import (
    handle_database_error,
    handle_integrity_error,
)
from flask_authx.interfaces.repository import IUsersRepository
from flask_authx.utils.result import Result


class SQLAlchemyUsersRepository(IUsersRepository):
    COLUMN_MAPPER = {
        "username": "username",
        "is_authenticated": "is_authenticated",
        "role": "role",
    }

    def __init__(self) -> None:
        super().__init__()
        self.UserModel = get_user_model()

    @handle_database_error
    def all(self) -> Result[list[User], DatabaseError]:
        models = get_instance().session.query(self.UserModel).all()
        return Result.ok([m.to_entity() for m in models])

    @handle_database_error
    def get_by_id(self, id: int) -> Result[User, NotFoundError | DatabaseError]:
        model = get_instance().session.get(self.UserModel, id)
        if not model:
            return Result.fail(NotFoundError(f"User with id: {id} doesn't exist."))
        return Result.ok(model.to_entity())

    @handle_database_error
    def get_by_name(self, username: str) -> Result[User, NotFoundError | DatabaseError]:
        model = (
            get_instance()
            .session.query(self.UserModel)
            .filter(self.UserModel.username == username)
            .first()
        )

        if not model:
            return Result.fail(
                NotFoundError(f"User with username: {username} not found.")
            )

        return Result.ok(model.to_entity())

    @handle_database_error
    @handle_integrity_error
    def add(
        self, new: UserForm
    ) -> Result[User, ConflictError | ValidationError | DatabaseError]:
        model = self.UserModel.create(new)
        get_instance().session.add(model)
        get_instance().session.commit()
        return Result.ok(model.to_entity())

    @handle_database_error
    def remove(self, id: int) -> Result[None, NotFoundError | DatabaseError]:
        model = get_instance().session.get(self.UserModel, id)
        if not model:
            return Result.fail(NotFoundError(f"User with id: {id} not found."))
        get_instance().session.delete(model)
        get_instance().session.commit()
        return Result.ok(None)

    @handle_database_error
    @handle_integrity_error
    def update(
        self, id: int, values: UserUpdateData
    ) -> Result[User, ConflictError | ValidationError | NotFoundError | DatabaseError]:
        model = get_instance().session.get(self.UserModel, id)
        if not model:
            return Result.fail(NotFoundError(f"User with id: {id} not found."))

        for k, v in values.items():
            if k in self.COLUMN_MAPPER:
                setattr(model, self.COLUMN_MAPPER[k], v)

        get_instance().session.commit()
        return Result.ok(model.to_entity())



from enum import Enum

from flask_authx.errors import ProgrammingError


class Role(Enum):
    user = "user"
    """
    Authenticated user
    """

    admin = "admin"
    """
    Admin account
    """

    dev = "dev"
    """
    Developer account
    """

    unauth = "unauth"
    """
    Unauthenticated user
    """

    @staticmethod
    def parse(s: str) -> "Role":
        if not Role.is_parseable(s):
            raise ProgrammingError(f"Option '{s}' is not a member of the role enum.")

        val = getattr(Role, s)

        return val

    @staticmethod
    def is_parseable(s: str) -> bool:
        return hasattr(Role, s)

    def dump(self) -> str:
        return self.value



from datetime import datetime
from sqlalchemy import (
    Boolean,
    Column,
    Enum,
    ForeignKey,
    Integer,
    String,
    DateTime,
)
from sqlalchemy.orm import relationship

from flask_authx.entities import Session, User
from flask_authx.forms import SessionForm, UserForm
from flask_authx.role import Role


def create_models(db):
    """Create models dynamically using the provided SQLAlchemy instance."""

    class UserModel(db.Model):
        __tablename__ = "Users"

        id = Column(Integer, primary_key=True, autoincrement=True)
        username = Column(String(150), nullable=False, unique=True)
        password_hash = Column(String, nullable=False)
        is_authenticated = Column(Boolean, nullable=False)
        role = Column(Enum(Role), nullable=False)
        created_at = Column(DateTime, default=datetime.now)

        session = relationship(
            "SessionModel",
            uselist=False,
            back_populates="user",
            cascade="all, delete-orphan",
        )

        def to_entity(self) -> User:
            return User(
                self.id,
                self.username,
                self.password_hash,
                self.role,
                self.is_authenticated,
                self.created_at,
            )

        @staticmethod
        def create(u: UserForm):
            return UserModel(
                username=u.username,
                password_hash=u.password,
                role=u.role,
                is_authenticated=False,
            )

    class SessionModel(db.Model):
        __tablename__ = "Sessions"

        token = Column(String, primary_key=True, nullable=False)
        user_id = Column(Integer, ForeignKey("Users.id"), unique=True)
        created_at = Column(DateTime, default=datetime.now)

        user = relationship(UserModel, uselist=False, back_populates="session")

        def to_entity(self) -> Session:
            return Session(self.token, self.user, self.created_at)

        @staticmethod
        def create(s: SessionForm):
            return SessionModel(token=s.token, user_id=s.user.id)

    return UserModel, SessionModel


# Cache for dynamically created models
_cached = {}


def get_user_model():
    from flask_authx.database.sqlalchemy.shared.instance import get_instance

    db = get_instance()
    if id(db) not in _cached:
        _cached[id(db)] = create_models(db)
    return _cached[id(db)][0]


def get_session_model():
    from flask_authx.database.sqlalchemy.shared.instance import get_instance

    db = get_instance()
    if id(db) not in _cached:
        _cached[id(db)] = create_models(db)
    return _cached[id(db)][1]


# For backward compatibility - will be replaced at runtime
UserModel = None
SessionModel = None



from functools import wraps
from flask import request

from flask_authx.utils.responses import (
    MissingJsonFieldsResponse,
    JSONRequiredResponse,
    MissingQueryParamsResponse,
)


def json_fields(*, required: list[str] = [], optional: list[str] = []):
    def decorator(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            if not request.is_json:
                return JSONRequiredResponse.create()

            json: dict = request.json  # type: ignore
            missing: list[str] = []

            for fld in required + optional:
                if fld in required and fld not in json:
                    missing.append(fld)
                    continue
                kwargs[fld] = json[fld]

            if len(missing) != 0:
                return MissingJsonFieldsResponse.create(*missing)

            return f(*args, **kwargs)

        return wrapper

    return decorator


def query_fields(*, required: list[str] = [], optional: list[str] = []):
    def decorator(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            args = request.args
            missing: list[str] = []

            for arg in required + optional:
                if arg in required and arg not in args:
                    missing.append(arg)
                    continue
                kwargs[arg] = args[arg]

            if len(missing) != 0:
                return MissingQueryParamsResponse.create(*missing)

            return f(*args, **kwargs)

        return wrapper

    return decorator



from typing import Optional
from flask import Flask

from flask_authx.database.sqlalchemy.shared.instance import (
    set_sqlalchemy,
)
from flask_authx.errors import ProgrammingError
from flask_authx.interfaces.database import IDatabaseSetup
from flask_authx.interfaces.repository import ISessionsRepository, IUsersRepository
from flask_authx.interfaces.security import IPasswordHashing
from flask_authx.interfaces.service import IAuthService
from flask_authx.routes.users import UsersRoutes
from flask_authx.routes.auth import AuthRoutes
from flask_authx.services.password_hashing import BcryptPasswordHashing
from flask_authx.services.auth import AuthService
from flask_authx.config import load_config
from flask_authx.services.users import IUsersService, UsersService


class AuthX:
    def __init__(
        self,
        builder: "AuthXBuilder",
    ) -> None:
        self.app = builder.app
        self.users_repository = builder.users_repository
        self.sessions_repository = builder.sessions_repository
        self.auth_service = builder.auth_service
        self.database_setup = builder.database_setup
        self.auth_routes_prefix = builder.auth_routes_prefix
        self.users_routes_prefix = builder.users_routes_prefix
        self.password_hashing = builder.password_hashing
        self.users_service = builder.users_service

        if self.app:
            self.init_app(self.app)

    def init_app(self, app: Flask):
        authx: Optional[AuthX] = app.extensions.get("authx")

        if authx:
            raise RuntimeError(
                "A 'AuthX' instance has been detected in this app, use that instead (with app['authx'])."
            )

        app.extensions["authx"] = self

        load_config(app.config)

        self.password_hashing = self.password_hashing or BcryptPasswordHashing()

        if (
            not self.database_setup
            or not self.users_repository
            or not self.sessions_repository
        ):
            set_sqlalchemy(app)

            from flask_authx.database.sqlalchemy.repositories.session import (
                SQLAlchemySessionsRepository,
            )
            from flask_authx.database.sqlalchemy.repositories.users import (
                SQLAlchemyUsersRepository,
            )
            from flask_authx.database.sqlalchemy.setup import SQLAlchemyDatabaseSetup

            self.users_repository = SQLAlchemyUsersRepository()
            self.users_service = self.users_service or UsersService(
                self.users_repository, self.password_hashing
            )
            self.database_setup = SQLAlchemyDatabaseSetup(app, self.users_service)
            self.sessions_repository = SQLAlchemySessionsRepository()

        self.users_service = self.users_service or UsersService(
            self.users_repository, self.password_hashing
        )

        self.auth_service = self.auth_service or AuthService(
            self.sessions_repository, self.users_service, self.password_hashing
        )

        self.database_setup.init()

        UsersRoutes(
            self.users_service,
            self.sessions_repository,
            prefix=self.users_routes_prefix,
            app=app,
        )

        AuthRoutes(
            self.auth_service,
            self.sessions_repository,
            prefix=self.auth_routes_prefix,
            app=app,
        )


class AuthXBuilder:
    def __init__(self, app: Flask):
        self.app = app
        self.auth_routes_prefix: str = "/auth"
        self.users_routes_prefix: str = "/users"

        self.users_repository: Optional[IUsersRepository] = None
        self.sessions_repository: Optional[ISessionsRepository] = None
        self.database_setup: Optional[IDatabaseSetup] = None
        self.auth_service: Optional[IAuthService] = None
        self.password_hashing: Optional[IPasswordHashing] = None
        self.users_service: Optional[IUsersService] = None

    def set_users_repository(self, repository: IUsersRepository):
        self.users_repository = repository
        return self

    def set_sessions_repository(self, repository: ISessionsRepository):
        self.sessions_repository = repository
        return self

    def set_auth_service(self, service: IAuthService):
        self.auth_service = service
        return self

    def set_database_setup(self, setup: IDatabaseSetup):
        self.database_setup = setup
        return self

    def set_auth_prefix(self, prefix: str):
        if not prefix.startswith("/"):
            raise ProgrammingError("Routes prefixes must start with slash ('/')")
        self.auth_routes_prefix = prefix
        return self

    def set_users_prefix(self, prefix: str):
        if not prefix.startswith("/"):
            raise ProgrammingError("Routes prefixes must start with slash ('/')")
        self.users_routes_prefix = prefix
        return self

    def set_password_hashing(self, component: IPasswordHashing):
        self.password_hashing = component
        return self

    def set_users_service(self, component: IUsersService):
        self.users_service = component
        return self

    def build(self) -> AuthX:
        return AuthX(self)



from functools import wraps
from flask import request

from flask_authx.errors import DatabaseError, NotFoundError
from flask_authx.interfaces.repository import ISessionsRepository
from flask_authx.utils.responses import (
    InternalErrorResponse,
    RequireAuthenticationResponse,
)


def require_authentication(session_repository: ISessionsRepository):
    def decorator(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            header = request.headers.get("Authorization")

            if not header or not header.startswith("Bearer "):
                return RequireAuthenticationResponse.create()

            parts = header.split(" ")
            if len(parts) != 2:
                return RequireAuthenticationResponse.create()

            token = parts[1]

            # This means that the client is not authenticated
            if not token:
                return RequireAuthenticationResponse.create()

            res = session_repository.get_by_token(token)

            if not res.success:
                if res.error_is(NotFoundError):
                    return RequireAuthenticationResponse.create()

                elif res.error_is(DatabaseError):
                    return InternalErrorResponse.create(res.error.message)

            return f(*args, **kwargs, user=res.value.user)

        return wrapper

    return decorator



from dataclasses import dataclass
from datetime import datetime

from flask_authx.errors import ValidationError
from flask_authx.role import Role
from flask_authx.utils.result import Result


@dataclass
class User:
    id: int
    username: str
    password: str
    """
    Plain or hashed password.
    """

    role: Role
    is_authenticated: bool
    created_at: datetime

    @classmethod
    def validate_username(cls, username: str) -> Result[None, ValidationError]:
        if len(username) < 3:
            return Result.fail(
                ValidationError("Username must contains at least 3 characters.")
            )
        return Result.ok(None)

    @classmethod
    def validate_password(cls, password: str) -> Result[None, ValidationError]:
        if len(password) < 8:
            return Result.fail(
                ValidationError("Password must contains at least 8 characters.")
            )
        return Result.ok(None)


@dataclass
class Session:
    token: str
    user: User
    created_at: datetime



from flask_authx.extension import AuthX as AuthX
from flask_authx.extension import AuthXBuilder as AuthXBuilder
from flask_authx.interfaces.database import IDatabaseSetup as IDatabaseSetup
from flask_authx.interfaces.repository import ISessionsRepository as ISessionsRepository
from flask_authx.interfaces.repository import IUsersRepository as IUsersRepository
from flask_authx.interfaces.service import IAuthService as IAuthService
from flask_authx.interfaces.security import IPasswordHashing



from flask_authx.entities import User
from flask_authx.errors import (
    ConflictError,
    DatabaseError,
    NotFoundError,
    ValidationError,
)
from flask_authx.forms import UserForm, UserUpdateData
from flask_authx.interfaces.service import IUsersService
from flask_authx.interfaces.repository import IUsersRepository
from flask_authx.interfaces.security import IPasswordHashing
from flask_authx.role import Role
from flask_authx.utils.result import Result


class UsersService(IUsersService):
    def __init__(
        self, users_repository: IUsersRepository, password_hashing: IPasswordHashing
    ) -> None:
        self.users_repository = users_repository
        self.password_hashing = password_hashing

    def all(self) -> Result[list[User], DatabaseError]:
        res = self.users_repository.all()

        if not res.success:
            return Result.fail(DatabaseError(res.error.message))

        return Result.ok(res.value)

    def get_by_id(self, id: int) -> Result[User, NotFoundError | DatabaseError]:
        res = self.users_repository.get_by_id(id)

        if not res.success:
            if res.error_is(NotFoundError):
                return Result.fail(NotFoundError(f"User with id {id} not found"))

            return Result.fail(DatabaseError(res.error.message))

        return Result.ok(res.value)

    def get_by_name(self, username: str) -> Result[User, NotFoundError | DatabaseError]:
        res = self.users_repository.get_by_name(username)

        if not res.success:
            if res.error_is(NotFoundError):
                return Result.fail(
                    NotFoundError(f"User with name {username} not found")
                )

            return Result.fail(DatabaseError(res.error.message))

        return Result.ok(res.value)

    def add(
        self, form: UserForm
    ) -> Result[User, ConflictError | ValidationError | DatabaseError]:
        if not form.is_valid().success:
            return Result.fail(ValidationError("Invalid user form data"))

        existing_user_res = self.users_repository.get_by_name(form.username)

        if existing_user_res.success:
            return Result.fail(
                ConflictError(f"User with name {form.username} already exists")
            )

        hashed_password = self.password_hashing.hash(form.password)
        new_form = UserForm(
            username=form.username, password=hashed_password, role=form.role
        )

        add_res = self.users_repository.add(new_form)

        if not add_res.success:
            return Result.fail(DatabaseError(add_res.error.message))

        return Result.ok(add_res.value)

    def remove(self, id: int) -> Result[None, NotFoundError | DatabaseError]:
        res = self.users_repository.remove(id)

        if not res.success:
            if res.error_is(NotFoundError):
                return Result.fail(NotFoundError(f"User with id {id} not found"))
            return Result.fail(DatabaseError(res.error.message))

        return Result.ok(None)

    def update_name(
        self, id: int, new_username: str
    ) -> Result[User, ValidationError | NotFoundError | ConflictError | DatabaseError]:
        valid_name_res = User.validate_username(new_username)

        if not valid_name_res.success:
            return Result.fail(
                ValidationError(
                    f"Invalid username. INFO: {valid_name_res.error.message}"
                )
            )

        existing_user_res = self.users_repository.get_by_name(new_username)

        if existing_user_res.success:
            return Result.fail(
                ConflictError(f"User with name {new_username} already exists")
            )

        else:
            if existing_user_res.error_is(DatabaseError):
                return Result.fail(DatabaseError(existing_user_res.error.message))

        update_data: UserUpdateData = {"username": new_username}
        update_res = self.users_repository.update(id, update_data)

        if not update_res.success:
            if update_res.error_is(NotFoundError):
                return Result.fail(NotFoundError(f"User with id {id} not found"))

            return Result.fail(DatabaseError(update_res.error.message))

        return Result.ok(update_res.value)

    def update_role(
        self, id: int, role: Role
    ) -> Result[User, NotFoundError | DatabaseError]:
        update_data: UserUpdateData = {"role": role}
        update_res = self.users_repository.update(id, update_data)

        if not update_res.success:
            if update_res.error_is(NotFoundError):
                return Result.fail(NotFoundError(f"User with id {id} not found"))

            return Result.fail(DatabaseError("Cannot update user role"))

        return Result.ok(update_res.value)

    def mark_authenticated(
        self, id: int
    ) -> Result[User, NotFoundError | DatabaseError]:
        update_data: UserUpdateData = {"is_authenticated": True}
        update_res = self.users_repository.update(id, update_data)

        if not update_res.success:
            if update_res.error_is(NotFoundError):
                return Result.fail(NotFoundError(f"User with id {id} not found"))

            return Result.fail(DatabaseError("Cannot mark user as authenticated"))

        return Result.ok(update_res.value)

    def mark_unauthenticated(
        self, id: int
    ) -> Result[User, NotFoundError | DatabaseError]:
        update_data: UserUpdateData = {"is_authenticated": False}
        update_res = self.users_repository.update(id, update_data)

        if not update_res.success:
            if update_res.error_is(NotFoundError):
                return Result.fail(NotFoundError(f"User with id {id} not found"))

            return Result.fail(DatabaseError("Cannot mark user as unauthenticated"))

        return Result.ok(update_res.value)

    def is_authenticated(self, id: int) -> Result[bool, NotFoundError | DatabaseError]:
        res = self.users_repository.get_by_id(id)

        if not res.success:
            if res.error_is(NotFoundError):
                return Result.fail(NotFoundError(f"User with id {id} not found"))

            return Result.fail(DatabaseError(res.error.message))

        return Result.ok(res.value.is_authenticated)



from dataclasses import dataclass

from flask_authx.errors import ProgrammingError


@dataclass
class Result[T, E]:
    _success: bool
    _value: T | None
    _error: E | None

    @staticmethod
    def ok[U, F](value: U) -> "Result[U, F]":  # type: ignore
        return Result[U, F](True, value, None)

    @staticmethod
    def fail[U, F](error: F) -> "Result[U, F]":  # type: ignore
        return Result[U, F](False, None, error)

    def error_is(self, error_type: type[E]) -> bool:
        return isinstance(self._error, error_type)

    @property
    def value(self) -> T:
        if self._value is None:
            raise ProgrammingError(
                "Cannot obtain 'value' property from NOT successful result."
            )
        return self._value

    @property
    def error(self) -> E:
        if self._error is None:
            raise ProgrammingError(
                "Cannot obtain 'error' property from successful result."
            )
        return self._error

    @property
    def success(self) -> bool:
        return self._success



from flask import Config
from flask_authx.errors import ConfigurationError


FIRST_USER_NAME: str
FIRST_USER_PASSWORD: str


def load_config(app_config: Config):
    global FIRST_USER_NAME, FIRST_USER_PASSWORD

    try:
        FIRST_USER_NAME = app_config["FIRST_USER_NAME"]
        FIRST_USER_PASSWORD = app_config["FIRST_USER_PASSWORD"]

    except KeyError as field:
        raise ConfigurationError(f"Missing field in app config: {field}")



from flask import Flask
from flask_sqlalchemy import SQLAlchemy


instance: SQLAlchemy = None  # pyright: ignore[reportAssignmentType]


def set_sqlalchemy(app: Flask):
    sa = ensure_sqlalchemy(app)
    global instance
    instance = sa


def ensure_sqlalchemy(app: Flask) -> SQLAlchemy:
    if not app.extensions.get("sqlalchemy"):
        raise RuntimeError(
            "App must have initialized 'SQLAlchemy' instance, from `Flask-SQLAlchemy` extension. Or define your own database configuration via `AuthXBuilder`."
        )
    return app.extensions["sqlalchemy"]


def get_instance() -> SQLAlchemy:
    """Get the current SQLAlchemy instance. Always returns the current value."""
    return instance



from typing import Optional
from flask import Blueprint, Flask, url_for

from flask_authx.entities import User
from flask_authx.forms import UserForm
from flask_authx.role import Role
from flask_authx.errors import (
    ConflictError,
    DatabaseError,
    NotFoundError,
    ValidationError,
)
from flask_authx.interfaces.repository import ISessionsRepository
from flask_authx.interfaces.service import IUsersService
from flask_authx.security.auth import require_authentication
from flask_authx.utils.request import json_fields
from flask_authx.utils.responses import (
    BadRequestFormatResponse,
    CreatedUserResponse,
    InternalErrorResponse,
    UserFormValidationErrorResponse,
    SuccessfulEmptyResponse,
    PermissionErrorResponse,
    UserInfoResponse,
    UserNotFoundResponse,
    UsernameConflictResponse,
    UsersListResponse,
)


class UsersRoutes:
    def __init__(
        self,
        users_service: IUsersService,
        sessions_repository: ISessionsRepository,
        *,
        prefix: str,
        app: Optional[Flask] = None,
    ) -> None:
        self.users_service = users_service
        self.sessions_repository = sessions_repository

        self._users_bp = Blueprint("users", __name__, url_prefix=prefix)
        self._require_authentication = require_authentication(self.sessions_repository)
        self._setup_routes()

        if app:
            self.init_routes(app)

    def init_routes(self, app: Flask):
        app.register_blueprint(self._users_bp)

    def _setup_routes(self):
        @self._users_bp.route("", methods=["GET"])
        def get_all():
            # Tested
            res = self.users_service.all()

            if not res.success:
                if res.error_is(DatabaseError):
                    return InternalErrorResponse.create(res.error.message)

            return UsersListResponse.create(res.value)

        @self._users_bp.route("/<int:id>", methods=["GET"])
        def get_by_id(id: int):
            # Tested
            res = self.users_service.get_by_id(id)

            if not res.success:
                if res.error_is(NotFoundError):
                    return UserNotFoundResponse.create(id=id)

                elif res.error_is(DatabaseError):
                    return InternalErrorResponse.create(res.error.message)

            return UserInfoResponse.create(res.value)

        @self._users_bp.route("/<string:username>", methods=["GET"])
        def get_by_username(username: str):
            # Tested
            res = self.users_service.get_by_name(username)

            if not res.success:
                if res.error_is(NotFoundError):
                    return UserNotFoundResponse.create(username=username)

                elif res.error_is(DatabaseError):
                    return InternalErrorResponse.create(res.error.message)

            return UserInfoResponse.create(res.value)

        @self._users_bp.route("/<int:id>", methods=["DELETE"])
        @self._require_authentication
        def delete(id: int, user: User):
            # Tested
            if user.id != id and user.role != Role.admin:
                return PermissionErrorResponse.create()

            session_res = self.sessions_repository.remove_by_user(id)

            if not session_res.success:
                if session_res.error_is(DatabaseError):
                    return InternalErrorResponse.create(session_res.error.message)

            res = self.users_service.remove(id)

            if not res.success:
                if res.error_is(NotFoundError):
                    return UserNotFoundResponse.create(id=id)

                elif res.error_is(DatabaseError):
                    return InternalErrorResponse.create(res.error.message)

            return SuccessfulEmptyResponse.create()

        @self._users_bp.route("", methods=["POST"])
        @json_fields(required=["username", "password", "role"])
        def post(username: str, password: str, role: str):
            # Tested
            if not Role.is_parseable(role):
                return BadRequestFormatResponse.create(
                    "The value of the field 'role' isn't valid."
                )

            form = UserForm(username, password, Role.parse(role))
            valid_res = form.is_valid()

            if not valid_res.success:
                return UserFormValidationErrorResponse.create(valid_res.error.message)

            res = self.users_service.add(form)
            user = res.value

            if not res.success:
                if res.error_is(ConflictError):
                    return UsernameConflictResponse.create()

                elif res.error_is(ValidationError):
                    return InternalErrorResponse.create(
                        f"Unhandled validation error: {res.error.message}"
                    )

                elif res.error_is(DatabaseError):
                    return InternalErrorResponse.create(res.error.message)

            url = url_for("users.get_by_id", _method="GET", _external=True, id=user.id)

            return CreatedUserResponse.create(res.value, url)



from typing import Optional
from flask import Blueprint, Flask

from flask_authx.entities import User
from flask_authx.forms import UserForm
from flask_authx.role import Role
from flask_authx.errors import (
    ConflictError,
    InvalidCredentialsError,
    NotAuthenticatedError,
)
from flask_authx.interfaces.repository import ISessionsRepository
from flask_authx.interfaces.service import IAuthService
from flask_authx.security.auth import require_authentication
from flask_authx.utils.request import json_fields
from flask_authx.utils.responses import (
    InternalErrorResponse,
    InvalidCredentialsResponse,
    RequireAuthenticationResponse,
    SessionAlreadyOpenResponse,
    SuccessfulEmptyResponse,
    SuccessfulLoginResponse,
)


class AuthRoutes:
    def __init__(
        self,
        auth_service: IAuthService,
        sessions_repository: ISessionsRepository,
        *,
        prefix: str,
        app: Optional[Flask] = None,
    ) -> None:
        self.auth_svc = auth_service
        self.sessions_repository = sessions_repository

        self._auth_bp = Blueprint("auth", __name__, url_prefix=prefix)
        self._require_authentication = require_authentication(self.sessions_repository)
        self._setup_routes()

        if app:
            self.init_routes(app)

    def init_routes(self, app: Flask):
        app.register_blueprint(self._auth_bp)

    def _setup_routes(self):
        @self._auth_bp.route("/login", methods=["POST"])
        @json_fields(required=["username", "password"])
        def login(username: str, password: str):
            # Tested
            form = UserForm(username, password, Role.unauth)
            login_res = self.auth_svc.login(form)

            if not login_res.success:
                if login_res.error_is(InvalidCredentialsError):
                    return InvalidCredentialsResponse.create()

                elif login_res.error_is(ConflictError):
                    return SessionAlreadyOpenResponse.create()

                else:
                    return InternalErrorResponse.create(login_res.error.message)

            sess = login_res.value

            return SuccessfulLoginResponse.create(sess)

        @self._auth_bp.route("/logout", methods=["DELETE"])
        @self._require_authentication
        def logout(user: User):
            # Tested
            logout_res = self.auth_svc.logout(user)

            if not logout_res.success:
                if logout_res.error_is(NotAuthenticatedError):
                    return RequireAuthenticationResponse.create()

                else:
                    return InternalErrorResponse.create(logout_res.error.message)

            return SuccessfulEmptyResponse.create()
